"""A color picker gadget."""

from collections.abc import Callable

import numpy as np

from ..colors import (
    ABLACK,
    ABLUE,
    ACYAN,
    AGREEN,
    AMAGENTA,
    ARED,
    AWHITE,
    AYELLOW,
    RED,
    WHITE,
    AColor,
    Color,
    gradient,
)
from .behaviors.grabbable import Grabbable
from .behaviors.themable import Themable
from .button import Button
from .gadget import Gadget, Point, Pointlike, PosHint, Size, SizeHint, Sizelike
from .graphics import Graphics, scale_geometry
from .pane import Pane
from .text import Text, new_cell

__all__ = ["ColorPicker", "Point", "Size"]

GRADIENT = ARED, AYELLOW, AGREEN, ACYAN, ABLUE, AMAGENTA, ARED


class _ShadeSelector(Grabbable, Graphics):
    def __init__(self, color_swatch, label, **kwargs):
        self._shade_hint = 0.0, 1.0
        self._shade_indicator = Text(size=(1, 1), is_transparent=True, default_cell="○")
        self.hue = ARED
        self.color_swatch = color_swatch
        self.label = label

        super().__init__(**kwargs)
        self.add_gadget(self._shade_indicator)
        self.update_hue(ARED)

    def on_size(self):
        super().on_size()
        h, w = self._size
        hh, wh = self._shade_hint

        self._shade_indicator.pos = round((h - 1) * hh), round((w - 1) * wh)
        self.update_hue(self.hue)

    def update_hue(self, hue: AColor):
        self.hue = hue

        h, w = scale_geometry(self._blitter, self._size)
        if w == 0:
            return

        self.texture = np.linspace(
            np.linspace(AWHITE, hue, w),
            np.linspace(ABLACK, ABLACK, w),
            h,
            dtype=np.uint8,
        )
        self.update_swatch_label()

    def update_swatch_label(self):
        y, x = scale_geometry(self._blitter, self._shade_indicator.pos)
        r, g, b = self.texture[y, x, :3].tolist()
        self.color_swatch.bg_color = r, g, b
        self.label.add_str(hex(r * 2**16 + g * 2**8 + b)[2:], pos=(1, 1))
        self.label.add_str(f"R: {r:>3}", pos=(3, 1))
        self.label.add_str(f"G: {g:>3}", pos=(4, 1))
        self.label.add_str(f"B: {b:>3}", pos=(5, 1))

    def grab(self, mouse_event):
        super().grab(mouse_event)
        self.grab_update(mouse_event)

    def grab_update(self, mouse_event):
        if self.collides_point(mouse_event.pos):
            y, x = self._shade_indicator.pos = self.to_local(mouse_event.pos)
            h, w = self.size
            self._shade_hint = (
                0 if h <= 1 else y / (h - 1),
                0 if w <= 1 else x / (w - 1),
            )
            self.update_swatch_label()


class _HueSelector(Grabbable, Graphics):
    def __init__(self, shade_selector, **kwargs):
        self._hue_hint = 0.0
        self._hue_indicator = Text(
            size=(1, 1),
            default_cell=new_cell(ord=ord("▼"), fg_color=WHITE, bg_color=RED),
        )
        self.shade_selector = shade_selector
        super().__init__(**kwargs)
        self.add_gadget(self._hue_indicator)

    def on_size(self):
        super().on_size()
        self.texture[:] = gradient(*GRADIENT, n=self.width)
        self._hue_indicator.x = round(self._hue_hint * self.width)
        self.update_hue()

    def update_hue(self):
        x = self._hue_indicator.x
        self._hue_indicator.canvas["bg_color"] = self.texture[0, x, :3]
        self.shade_selector.update_hue(AColor(*self.texture[0, x]))

    def grab(self, mouse_event):
        super().grab(mouse_event)
        self.grab_update(mouse_event)

    def grab_update(self, mouse_event):
        if self.collides_point(mouse_event.pos):
            x = self._hue_indicator.x = self.to_local(mouse_event.pos).x
            self._hue_hint = 0 if self.width <= 1 else x / (self.width - 1)
            self.update_hue()


class ColorPicker(Themable, Gadget):
    r"""
    A color picker gadget.

    Parameters
    ----------
    ok_callback : Callable[[Color], None], default: lambda color: None
        Called with currently selected color when "OK" button is released.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    get_color(color_name)
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        ok_callback: Callable[[Color], None] = lambda color: None,
        alpha: float = 1.0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.color_swatch = Pane(pos=(1, 1), bg_color=RED, is_transparent=False)
        self.label = Text(size=(9, 8), alpha=0, is_transparent=True)
        self.shades = _ShadeSelector(
            pos=(1, 1),
            color_swatch=self.color_swatch,
            label=self.label,
            is_transparent=False,
            blitter="full",
        )
        self.hues = _HueSelector(
            pos=(1, 1), shade_selector=self.shades, is_transparent=False
        )
        ok_button = Button(
            label="OK",
            size=(1, 6),
            pos=(7, 1),
            callback=lambda: ok_callback(self.color_swatch.bg_color),
        )
        self._container = Pane(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
            is_transparent=is_transparent,
        )
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.alpha = alpha
        self.label.add_gadget(ok_button)
        self._container.add_gadgets(
            self.color_swatch, self.hues, self.shades, self.label
        )
        self.add_gadget(self._container)
        self.on_size()

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._container.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._container.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._container.is_transparent = self.is_transparent

    def on_size(self):
        """Resize and reposition children."""
        h, w = self._size

        self.shades.size = max(10, h - 4), max(20, w - 11)

        self.color_swatch.size = max(2, h - 12), 8
        self.color_swatch.left = self.shades.right + 1

        self.hues.size = 1, self.shades.width
        self.hues.top = self.shades.bottom + 1

        self.label.top = self.color_swatch.bottom + 1
        self.label.left = self.shades.right + 1

    def update_theme(self):
        """Paint the gadget with current theme."""
        primary_fg = self.get_color("primary_fg")
        primarg_bg = self.get_color("primary_bg")
        self._container.bg_color = primarg_bg
        self.label.default_fg_color = self.label.canvas["fg_color"] = primary_fg
        self.label.default_bg_color = self.label.canvas["bg_color"] = primarg_bg



"""A windows VT100 terminal."""

import asyncio
import msvcrt
from collections.abc import Callable
from ctypes import Structure, byref, windll
from ctypes.wintypes import BOOL, DWORD, HANDLE, LPVOID

from .events import Event
from .vt100_terminal import Vt100Terminal

# See: https://msdn.microsoft.com/pl-pl/library/windows/desktop/ms686033(v=vs.85).aspx
ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004
ENABLE_VIRTUAL_TERMINAL_INPUT = 0x0200


class SECURITY_ATTRIBUTES(Structure):
    """
    Windows security attributes.

    References
    ----------
    http://msdn.microsoft.com/en-us/library/windows/desktop/aa379560(v=vs.85).aspx
    """

    _fields_ = [
        ("nLength", DWORD),
        ("lpSecurityDescriptor", LPVOID),
        ("bInheritHandle", BOOL),
    ]


class WindowsTerminal(Vt100Terminal):
    """
    A windows VT100 terminal.

    Attributes
    ----------
    stdin: int
        The stdin file descriptor.
    stdout: int
        The stdout file descriptor.
    in_alternate_screen : bool
        Whether the alternate screen buffer is enabled.

    Methods
    -------
    process_stdin()
        Read from stdin and feed data into input parser to generate events.
    raw_mode()
        Set terminal to raw mode.
    restore_console()
        Restore console to its original mode.
    attach(event_handler)
        Start generating events from stdin.
    unattach()
        Stop generating events from stdin.
    feed(input_, reset_before)
        Write bytes to stdin parser and return generated events.
    events()
        Return a list of input events and reset the event buffer.
    get_size()
        Get terminal size.
    write(out)
        Write bytes directly to the out-buffer.
    flush()
        Write out-buffer to output stream and flush.
    set_title(title)
        Set terminal title.
    enter_alternate_screen()
        Enter alternate screen buffer.
    exit_alternate_screen()
        Exit alternate screen buffer.
    enable_mouse_support()
        Enable mouse support in terminal.
    disable_mouse_support()
        Disable mouse support in terminal.
    enable_sgr_pixels()
        Enable SGR-PIXELS mouse mode.
    reset_attributes()
        Reset character attributes.
    enable_bracketed_paste()
        Enable bracketed paste in terminal.
    disable_bracketed_paste()
        Disable bracketed paste in terminal.
    show_cursor()
        Show cursor in terminal.
    hide_cursor()
        Hide cursor in terminal.
    enable_reporting_focus()
        Enable reporting terminal focus.
    disable_reporting_focus()
        Disable reporting terminal focus.
    request_cursor_position_report()
        Report current cursor position.
    request_foreground_color()
        Report terminal foreground color.
    request_background_color()
        Report terminal background color.
    request_device_attributes()
        Report device attributes.
    request_pixel_geometry()
        Report pixel geometry per cell.
    request_terminal_geometry()
        Report pixel geometry of terminal.
    request_sgr_pixels_supported()
        Report whether SGR-PIXELS mouse mode is supported.
    request_synchronized_update_mode_supported()
        Report whether synchronized update mode is supported.
    expect_dsr()
        Return whether a device status report is expected.
    move_cursor(pos)
        Move cursor to ``pos``.
    erase_in_display(n)
        Clear part of the screen.
    """

    def __init__(self, stdin: int = 0, stdout: int = 1) -> None:
        super().__init__(stdin, stdout)
        self._original_input_mode = DWORD()
        """Original console input mode."""
        self._stdin_handle = msvcrt.get_osfhandle(stdin)
        self._stdout_handle = msvcrt.get_osfhandle(stdout)

        windll.kernel32.GetConsoleMode(
            self._stdin_handle, byref(self._original_input_mode)
        )

        self._original_output_mode = DWORD()
        """Original console output mode."""
        windll.kernel32.GetConsoleMode(
            self._stdout_handle, byref(self._original_output_mode)
        )

        self._original_input_cp = windll.kernel32.GetConsoleCP()
        """Original console input code page."""
        self._original_output_cp = windll.kernel32.GetConsoleOutputCP()
        """Original console output code page."""

    def raw_mode(self) -> None:
        """Set terminal to raw mode."""
        windll.kernel32.SetConsoleMode(
            self._stdin_handle, ENABLE_VIRTUAL_TERMINAL_INPUT
        )
        windll.kernel32.SetConsoleMode(
            self._stdout_handle,
            self._original_output_mode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING,
        )
        windll.kernel32.SetConsoleCP(65001)
        windll.kernel32.SetConsoleOutputCP(65001)

    def restore_console(self) -> None:
        """Restore console to its original mode."""
        windll.kernel32.SetConsoleMode(self._stdin_handle, self._original_input_mode)
        windll.kernel32.SetConsoleMode(self._stdout_handle, self._original_output_mode)
        windll.kernel32.SetConsoleCP(self._original_input_cp)
        windll.kernel32.SetConsoleOutputCP(self._original_output_cp)

    def attach(self, event_handler: Callable[[list[Event]], None]) -> None:
        """
        Start generating events from stdin.

        Parameters
        ----------
        event_handler : Callable[[list[Event]], None]
            Callable that handles input events.
        """
        self._event_buffer.clear()
        self._event_handler = event_handler
        self._remove_event = remove_event = HANDLE(
            windll.kernel32.CreateEventA(
                SECURITY_ATTRIBUTES(), BOOL(True), BOOL(False), None
            )
        )
        loop = asyncio.get_running_loop()
        wait_for = windll.kernel32.WaitForMultipleObjects
        EVENTS = (HANDLE * 2)(remove_event, self._stdin_handle)

        def ready():
            try:
                self.process_stdin()
                if self._event_handler is not None:
                    self._event_handler(self.events())
            finally:
                try:
                    loop.run_in_executor(None, wait)
                except RuntimeError:
                    windll.kernel32.CloseHandle(remove_event)

        def wait():
            if wait_for(2, EVENTS, BOOL(False), DWORD(-1)) == 0:
                windll.kernel32.CloseHandle(remove_event)
            else:
                loop.call_soon_threadsafe(ready)

        loop.run_in_executor(None, wait)

    def unattach(self) -> None:
        """Stop generating events from stdin."""
        self._event_handler = None
        windll.kernel32.SetEvent(self._remove_event)



"""Functions and classes for syntax highlighting with tree-sitter."""

from __future__ import annotations

import importlib
from contextlib import suppress
from dataclasses import dataclass
from functools import cache
from pathlib import Path

from tree_sitter import Language, Parser, Query, Range

from ..logging import get_logger

__all__ = ["Highlighter", "TSPoint", "get_highlighter", "register_language"]

logger = get_logger(__name__)

_REGISTERED_HIGHLIGHTERS: dict[str, Highlighter] = {}

type TSPoint = tuple[int, int]
"""A tree-sitter (row, byte_offset) tuple."""


@dataclass
class Highlighter:
    """Language, parser, and queries needed for syntax highlighting."""

    language_name: str
    """Name of language."""
    language: Language
    """A tree-sitter language."""
    parser: Parser
    """A tree-sitter parser for given language."""
    highlights: Query
    """A highlight query for syntax highlighting."""
    injections: Query | None = None
    """Optional injection query for injected languages."""


def register_language(
    language_name: str,
    language: Language,
    highlights_source: str,
    injections_source: str | None = None,
) -> None:
    """
    Register a tree-sitter language and its queries for syntax highlighting.

    Once registered, a ``Highlighter`` for ``language_name`` can be retrieved with
    ``get_highlighter``.

    Parameters
    ----------
    language_name : str
        Name of the language.
    language : tree_sitter.Language
        A tree-sitter language.
    highlights_source : str
        Source for highlight queries.
    injections_source : str | None, default: None
        Optional source for injection queries. Injection language must exist to be
        highlighted.
    """
    parser = Parser(language)
    highlights_query = Query(language, highlights_source)
    if injections_source is not None:
        injections_query = Query(language, injections_source)
    else:
        injections_query = None

    _REGISTERED_HIGHLIGHTERS[language_name] = Highlighter(
        language_name, language, parser, highlights_query, injections_query
    )
    get_highlighter.cache_clear()


def _find_tree_sitter_language(language: str) -> Language | None:
    with suppress(ImportError):
        tree_module = importlib.import_module(f"tree_sitter_{language}")
        _lang: object = tree_module.language()
        return Language(_lang)

    with suppress(ImportError):
        tree_module = importlib.import_module("tree_sitter_language_pack")
        with suppress(LookupError):
            return tree_module.get_language(language)


@cache
def get_highlighter(language: str) -> Highlighter | None:
    """
    Return a highlighter for a given language.

    Parameters
    ----------
    language : str
        Name of the language.

    Returns
    -------
    Highlighter | None
        Highlighter for a given language if found.
    """
    if language in _REGISTERED_HIGHLIGHTERS:
        return _REGISTERED_HIGHLIGHTERS[language]

    tree_lang = _find_tree_sitter_language(language)
    if tree_lang is None:
        logger.info("Could not load '%s' tree-sitter Language.", language)
        return None

    parser = Parser(tree_lang)

    queries_path = Path(__file__).parent / language
    highlights_path = queries_path / "highlights.scm"
    if not highlights_path.exists():
        logger.info(
            "Highlight query for '%s' language not found at '%s'.",
            language,
            highlights_path.resolve(),
        )
        return None
    highlights = Query(tree_lang, highlights_path.read_text())

    injections_path = queries_path / "injections.scm"
    if not injections_path.exists():
        injections = None
    else:
        injections = Query(tree_lang, injections_path.read_text())

    return Highlighter(language, tree_lang, parser, highlights, injections)


def changed_ranges_point_range(changed_ranges: list[Range]) -> tuple[TSPoint, TSPoint]:
    """
    Combine a syntax tree's changed ranges into a single range.

    Parameters
    ----------
    changed_ranges : list[Range]
        Non-empty list of changed ranges.

    Returns
    -------
    tuple[TSPoint, TSPoint]
        Combined range of changed ranges.
    """
    min_y = changed_ranges[0].start_point[0]
    max_y = changed_ranges[0].end_point[0]
    for range_ in changed_ranges:
        if range_.start_point[0] < min_y:
            min_y = range_.start_point[0]
        if range_.end_point[0] > max_y:
            max_y = range_.end_point[0]

    min_y -= 2
    if min_y < 0:
        min_y = 0

    return (min_y, 0), (max_y + 2, 0)
